{DelphiSim by Roeland van der Spek           }
{spekw@acm.org                               }
{(C) 1997                                    }
{v2.00                                       }
{Random distribution Streams and random      }
{functions                                   }


unit Random;

interface

uses  Windows, Messages, SysUtils,classes, Graphics, Controls, Forms, Dialogs,math,delphisim;

Const SeriesSize = 100;
      two2the26th = 67108864;

type

  TDistribution = class;

  TRandomStream = class(TComponent)
  private
    S1,S2:Longint;
    Mseed : longint;
    LSeed : longint;
    DistributionList:TQueue;
    Series: array[0..SeriesSize-1] of extended;
    Function MGen:extended;
  protected
    procedure loaded; override;
  public
    constructor create(owner:TComponent); override;
    destructor destroy; override;
    Procedure AddDistribution(D:TDistribution);
    Procedure RemoveDistribution(D:TDistribution);
    Function Uniform : extended;
    procedure Reset;
  published
    property Seed1:Longint read S1 write S1;
    property Seed2:Longint read S2 write S2;
  end;

  Tdistribution = class(TComponent)
  private
    U:TRandomStream;
  protected
    procedure loaded; override;
  public
    destructor destroy; override;
    function draw:extended; virtual; abstract;
    function uniform:extended;
  published
    property RandomStream:TrandomStream read U write U;
  end;

  TUniform = class(Tdistribution)
  private
    lo,hi,range:extended;
  public
    Function Draw:extended; override;
  published
    Property HighValue:Extended read hi write hi;
    property LowValue:Extended read lo write lo;
  end;


  TErlang = Class(TDistribution)
  private
    b:extended;
    param:longint;
  public
    Function Draw:extended; Override;
  published
    property Expomean:Extended read b write b;
    property Shape:Longint read param write param;
  end;


  TExponential = Class(TDistribution)
  private
    B:extended;
    procedure setB(m:Extended);
  public
    function draw:extended; override;
  published
    property mean:Extended read b write setB;
  end;

  TNormal = Class(TDistribution)
  private
    M,StDv,z:extended;
  public
    function draw:extended; override;
  published
    Constructor create(Owner:TComponent); override;
    property mean:extended read m write m;
    property Standarddeviation:Extended read Stdv write Stdv;
  end;


  TWeibull = Class(TDistribution)
  private
    A,B:Extended;
  public
    Function draw:Extended; override;
  Published
    property Beta:Extended read B write B;
    property Alpha:Extended read A write A;
  end;

  TTriangular = class(TDistribution)
  private
    min,modu,max:Extended;
  public
    function draw:Extended; override;
  published
    property minimum:Extended read min write min;
    property modus:Extended read modu write modu;
    property maximum:Extended read max write max;
  end;



function DrawNormal(m,s:extended;stream:TrandomStream):extended;
function DrawExpo(m:extended;Stream:TRandomStream):extended;
Function DrawErlang(m:extended;k:longint;Stream:TrandomStream):extended;
function drawUniform(l,u:extended;stream:TrandomStream):extended;
function drawWeibull(a,b:Extended;stream:TRandomStream):Extended;
function DrawTria(a,b,c:Extended;Stream:TrandomStream):extended;


procedure register;

implementation

{##############################################################################}
{############################## Random Streams ################################}
{##############################################################################}

Constructor TRandomStream.create(owner:TComponent);
begin
  inherited create(owner);
  distributionList:=TQueue.create;
end;

destructor TRandomStream.destroy;
var D:TDistribution;
begin
  while not DistributionList.IsEmpty do
  begin
    D:=DistributionList.pop;
    D.U:=nil;
  end;
  Distributionlist.free;
  inherited destroy;
end;

Procedure TRandomStream.AddDistribution(D:TDistribution);
var i:longint;
begin
  i:=Distributionlist.indexof(D);
  if i<0 then DistributionList.add(D);
end;

Procedure TRandomStream.RemoveDistribution(D:TDistribution);
var i:Longint;
begin
  i:=DistributionList.IndexOf(D);
  if I>=0 then DistributionList.Delete(i);
end;



Procedure TRandomStream.loaded;
begin
  Inherited loaded;
  reset;
end;

Procedure TRandomStream.Reset;
var i:longint;
begin
  if Seed1 mod 2 = 0 then Seed1:=Seed1 -1; {moet oneven zijn}
  if Seed1 <0 then Seed1:=-Seed1;
  if Seed2 <0 then Seed2:=-Seed2;
  MSeed:=S1;
  LSeed:=S2;
  for i:=0 to SeriesSize -1 do series[i]:=Mgen;
end;

Function TRandomStream.MGen:extended;
begin
  Mseed:=(Mseed*25) mod two2the26th;
  Mseed:=(Mseed*25) mod two2the26th;
  Mseed:=(Mseed*5) mod two2the26th;
  result:= Mseed/two2the26th;
end;

Function TrandomStream.Uniform:extended;
const m =100000000;
      b =31415821;
      m1=10000;
var Choose,p0,p1,q0,q1:longint;
begin
  p0:=LSeed mod m1;
  p1:=Lseed div m1;
  q0:=b mod m1;
  q1:=b div m1;

  Lseed := (((((p0*q1+p1*q0)mod m1)*m1+p0*q0)mod m) +1) mod m;
  Choose:=LSeed mod SeriesSize;
  Result := series[choose];
  Series[choose]:=Mgen;
end;

{##############################################################################}


function TDistribution.uniform:extended;
begin
  result:=u.uniform;
end;

destructor TDistribution.destroy;
begin
  if U<>nil then U.RemoveDistribution(self);
  inherited destroy;
end;

procedure TDistribution.loaded;
begin
  inherited loaded;
  if U<>NIl then U.adddistribution(self);
end;

function TUniform.draw:extended;
begin
  result:=lo+(range*uniform);
end;

Function TErlang.draw;
var i:longint;
    temp:extended;
begin
  temp:=uniform;
  for i:=2 to param do
  begin
    temp:=temp*uniform;
  end;
  result:=-B * ln(temp);
end;


Function TExponential.draw:extended;
begin
  result:=-B*ln(Uniform);
end;

Procedure TExponential.setB(m:Extended);
begin
  if m>0 then B:=m else B:=-m;
end;

constructor TNormal.create(Owner:TComponent);
begin
  inherited create(Owner);
  Z:=0;
end;

function TNormal.draw:extended;
var X2,S,V1,V2,temp:extended;
begin
  result:=0;
  try
    S:=1;
    v1:=0;
    v2:=0;
    if z<>0 then
    begin
      X2:=z;
      z:=0;
    end
    else
    begin
      while (S >= 1) do
      begin
        v1 := 2*uniform-1;
        v2 := 2*uniform-1;
        S:= v1*v1 +v2*v2;
      end;
      temp:=(-2*ln(s)/S);
      s:=sqrt(temp);
      X2:= v1*S;
      z:=v2*S;
    end;
    result:= mean+X2*Stdv;
  except
  end;
end;


Function TWeibull.draw;
begin
  result:=a * power(-ln( Uniform), (1/b))
end;


Function TTriangular.draw;
var C,X1,X2,S,T:Extended;
begin
  S:= modus - min;
  T:= Max - Min;
  C:= S/T;
  X1:= uniform;
  X2:= Uniform;
  if X1 <=C then result:= S* sqrt(X2)+min
  else result:= T-(T-S)*Sqrt(X2)+min;
end;



{#############################################################################}

function DrawNormal(m,s:extended;stream:TrandomStream):extended;
var N:TNormal;
begin
  N:=TNormal.create(Nil);
  N.mean:=m;
  N.StandardDeviation:=S;
  N.RandomStream:=Stream;
  result:=N.draw;
  N.free;
end;

function DrawExpo(m:extended;Stream:TRandomStream):extended;
var E:TExponential;
begin
  E:=TExponential.create(nil);
  E.Mean:=m;
  E.RandomStream:=Stream;
  result:=E.draw;
  E.free;
end;

Function DrawErlang(m:extended;k:longint;Stream:TrandomStream):extended;
var E:TErlang;
begin
  E:=Terlang.create(nil);
  E.shape:=K;
  E.Expomean:=m;
  E.RandomStream:=Stream;
  result:=E.draw;
  E.free;
end;

function drawUniform(l,u:extended;stream:TrandomStream):extended;
var Unif:TUniform;
begin
  Unif:=TUniform.create(nil);
  Unif.highvalue:=u;
  Unif.lowValue:=l;
  Unif.RandomStream:=Stream;
  Result:=Unif.draw;
  Unif.free;
end;

function drawWeibull(a,b:Extended;stream:TRandomStream):Extended;
var weib:TWeibull;
begin
  Weib:=TWeibull.create(nil);
  Weib.Alpha:=a;
  Weib.Beta:=b;
  Weib.RandomStream:=Stream;
  result:=weib.draw;
  weib.free;
end;

Function DrawTria(a,b,c:Extended;Stream:TrandomStream):Extended;
var tria:TTriangular;
begin
  Tria:=TTriangular.create(NIl);
  Tria.minimum:=a;
  Tria.Modus:=b;
  tria.maximum:=c;
  tria.randomstream:=stream;
  result:=tria.draw;
  tria.free;
end;

procedure register;
begin
  RegisterComponents('Random',[TRandomStream,TUniform,TWeibull,TNormal,TTriangular,TExponential,TErlang]);
end;

end.

